
//**************************************************************************
//**
//** $Id: sys_sock.c,v 1.1 2003/03/09 16:03:51 skyjake Exp $
//**
//** Handling of sockets.
//**
//** $Log: sys_sock.c,v $
//** Revision 1.1  2003/03/09 16:03:51  skyjake
//** New master server mechanism
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include <winsock.h>
#include <stdarg.h>
#include <stdio.h>

#include "sys_sock.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

/*
 * N_SockInit
 *	Called from N_Init().
 */
void N_SockInit(void)
{
	WSADATA wsaData;
	WSAStartup( MAKEWORD(1,1), &wsaData );
	// FIXME: Check the result... (who cares?)
}

/*
 * N_SockShutdown
 *	Called from N_Shutdown().
 */
void N_SockShutdown(void)
{
	WSACleanup();
}

/*
 * N_SockPrintf
 *	Don't print too long messages with one call.
 */
void N_SockPrintf(socket_t s, const char *format, ...)
{
	char buf[512];
	int length;
	va_list args;

	// Print the message into the buffer.
	va_start(args, format);
	length = vsprintf(buf, format, args);
	va_end(args);

	if(length > sizeof(buf)) 
	{
		// Oops... Something important may have been overwritten in memory.
		length = sizeof(buf);
	}

	// Send it.
	send(s, buf, length, 0);
}

/*
 * N_SockGetHost
 */
struct hostent *N_SockGetHost(const char *hostName)
{
	struct sockaddr_in addr;
	struct hostent *host;

	// Clear the socket address.
    memset(&addr, 0, sizeof(addr));

	// Is it an IP address or name?
    addr.sin_addr.s_addr = inet_addr(hostName);

	// Get host information.
    if(addr.sin_addr.s_addr == INADDR_NONE) 
	{
		// Try to get using DNS.
        if((host = gethostbyname(hostName)) == NULL) 
		{
            //perror("Could not get host by name");
            return NULL;
        }
    } 
	else 
	{
		// Try to get using the IP address.
        if((host = gethostbyaddr((const char*)&addr.sin_addr, 
			sizeof(struct sockaddr_in), AF_INET)) == NULL)
        {
            //perror("Could not get host by address");
            return NULL;
        }
    }
	return host;
}

/*
 * N_SockNewStream
 */
socket_t N_SockNewStream(void)
{
	return socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
}

/*
 * N_SockConnect
 */
boolean N_SockConnect(socket_t s, struct hostent *host, unsigned short port)
{
	struct sockaddr_in addr;

	// The address to connect to.
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr   = *((struct in_addr*) *host->h_addr_list);
	// 80 is the default port.
	addr.sin_port   = htons(port);

	// Let's try connecting.
	return connect(s, (struct sockaddr*) &addr, sizeof(struct sockaddr)) 
		!= SOCKET_ERROR;
}

/*
 * N_SockClose
 */
void N_SockClose(socket_t s)
{
	closesocket(s);
}
